from typing import Any, Dict, Optional
from uuid import UUID

from pydantic import BaseModel, field_validator
from weaviate.collections.classes.internal import MetadataReturn
from weaviate.outputs.query import Object, QueryReturn

from weaviate_agents.classes.core import Usage


class PersonalizedObject(BaseModel):
    uuid: UUID
    original_rank: int
    personalized_rank: Optional[int]
    properties: Dict[str, Any]

    def model_dump(self, **kwargs):
        data = super().model_dump(**kwargs)
        if "uuid" in data:
            data["uuid"] = str(data["uuid"])
        return data


class PersonalizationAgentGetObjectsResponse(BaseModel):
    objects: list[PersonalizedObject]
    ranking_rationale: Optional[str] = None
    usage: Usage


class PersonalizedQueryResponse(BaseModel, QueryReturn):
    usage: Usage

    @field_validator("objects", mode="after")
    @classmethod
    def _ensure_metadata_type(cls, objects: list[Object]) -> list[Object]:
        for i in range(len(objects)):
            if isinstance(existing_metadata := objects[i].metadata, dict):
                objects[i].metadata = MetadataReturn(**existing_metadata)
        return objects



from datetime import datetime
from typing import Any, Dict, Optional
from uuid import UUID

from pydantic import BaseModel, Field


class Persona(BaseModel):
    persona_id: UUID
    properties: Dict[str, Any]


class PersonaInteraction(BaseModel):
    """Interaction between a persona and an item.

    If `replace_previous_interactions` is `True`, the interaction history for that specific
    item and persona pair will be replaced with the new interaction.
    """

    persona_id: UUID
    item_id: UUID
    weight: float
    replace_previous_interactions: bool = False
    created_at: Optional[datetime] = None


class PersonaInteractionResponse(BaseModel):
    """Response model for persona interactions when retrieving them."""

    item_id: UUID = Field(alias="uuid")
    weight: float
    created_at: str = Field(alias="createdAt")



from typing import Any, Dict, List, Union

import httpx
import pytest
from weaviate.collections.classes.config import DataType

from weaviate_agents.transformation.classes import (
    AppendPropertyOperation,
    OperationType,
    TransformationResponse,
    UpdatePropertyOperation,
)
from weaviate_agents.transformation.transformation_agent import TransformationAgent


class DummyClient:
    """A dummy client to simulate the Weaviate client for testing purposes."""

    def __init__(self):
        self._connection = self  # Simulate connection
        self.url = "http://test-url:443"  # Add url attribute
        self.additional_headers = {"Authorization": "test-token"}

    def get_current_bearer_token(self) -> str:
        """Simulate bearer token retrieval from connection.

        Returns:
            str: The test bearer token prefixed with 'Bearer '.
        """
        return "Bearer test-token"


class FakeResponse:
    """A fake HTTP response for testing purposes.

    Attributes:
        status_code (int): The HTTP status code.
        _json (Union[Dict[str, Any], List[Dict[str, Any]]]): The JSON data to return.
    """

    def __init__(
        self, status_code: int, json_data: Union[Dict[str, Any], List[Dict[str, Any]]]
    ):
        self.status_code = status_code
        self.is_error = 400 <= status_code <= 599
        self._json = json_data
        self.text = json_data

    def json(self) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Return the JSON data of the fake response.

        Returns:
            Union[Dict[str, Any], List[Dict[str, Any]]]: The JSON data.
        """
        return self._json


def fake_post_success(*args, **kwargs) -> FakeResponse:
    """Simulate a successful HTTP POST response.

    Returns:
        FakeResponse: A fake HTTP response with status code 202.
    """
    json_data = {
        "workflow_id": "test_workflow_id",
        "status": "running",
    }
    return FakeResponse(202, json_data)


def fake_post_failure(*args, **kwargs) -> FakeResponse:
    """Simulate a failed HTTP POST response.

    Returns:
        FakeResponse: A fake HTTP response with a non-200 status code.
    """
    json_data = {
        "error": {
            "message": "Test error message",
            "code": "test_error_code",
            "details": {"info": "test detail"},
        }
    }
    return FakeResponse(400, json_data)


def fake_get_status_success(*args, **kwargs) -> FakeResponse:
    """Simulate a successful HTTP GET response for status check.

    Returns:
        FakeResponse: A fake HTTP response with status code 200.
    """
    json_data = {
        "workflow_id": "workflow1",
        "status": {
            "batch_count": 40,
            "end_time": "2024-03-05 14:23:39",
            "start_time": "2024-03-05 14:21:15",
            "state": "completed",
            "total_duration": 144.191817,
            "total_items": 10000,
        },
    }
    return FakeResponse(200, json_data)


def test_update_all_success(monkeypatch):
    """Test that TransformationAgent.update_all returns valid responses when HTTP call is successful.

    Returns:
        None.
    """
    monkeypatch.setattr(httpx.Client, "post", fake_post_success)
    dummy_client = DummyClient()

    operations = [
        AppendPropertyOperation(
            property_name="test_prop1",
            instruction="test instruction 1",
            view_properties=["prop1", "prop2"],
            data_type=DataType.TEXT,
        ),
        UpdatePropertyOperation(
            property_name="test_prop2",
            instruction="test instruction 2",
            view_properties=["prop3", "prop4"],
        ),
    ]

    agent = TransformationAgent(
        dummy_client,
        collection="test_collection",
        operations=operations,
        agents_host="http://dummy-agent",
    )

    result = agent.update_all()
    assert isinstance(result, TransformationResponse)
    assert result.workflow_id == "test_workflow_id"


def test_update_all_failure(monkeypatch):
    """Test that TransformationAgent.update_all raises an exception when HTTP response indicates an error.

    Returns:
        None.
    """
    monkeypatch.setattr(httpx.Client, "post", fake_post_failure)
    dummy_client = DummyClient()

    operations = [
        AppendPropertyOperation(
            property_name="test_prop",
            instruction="test instruction",
            view_properties=["prop1"],
            data_type=DataType.TEXT,
        ),
    ]

    agent = TransformationAgent(
        dummy_client,
        collection="test_collection",
        operations=operations,
        agents_host="http://dummy-agent",
    )

    with pytest.raises(Exception):
        agent.update_all()


def test_get_status_success(monkeypatch):
    """Test that TransformationAgent.get_status returns valid response when HTTP call is successful.

    Returns:
        None.
    """
    monkeypatch.setattr(httpx.Client, "get", fake_get_status_success)
    dummy_client = DummyClient()

    agent = TransformationAgent(
        dummy_client,
        collection="test_collection",
        operations=[],
        agents_host="http://dummy-agent",
    )

    result = agent.get_status("workflow1")
    assert isinstance(result, dict)
    assert result["workflow_id"] == "workflow1"
    assert isinstance(result["status"], dict)
    assert result["status"]["batch_count"] == 40
    assert result["status"]["state"] == "completed"
    assert result["status"]["total_items"] == 10000
    assert isinstance(result["status"]["total_duration"], float)
    assert "start_time" in result["status"]
    assert "end_time" in result["status"]


def test_invalid_operation_type():
    """Test that TransformationAgent validates operation types correctly.

    Returns:
        None.
    """
    dummy_client = DummyClient()

    # Create an operation with mismatched type
    operation = UpdatePropertyOperation(
        property_name="test_prop",
        instruction="test instruction",
        view_properties=["prop1"],
    )
    operation.operation_type = OperationType.APPEND  # Force wrong type

    agent = TransformationAgent(
        dummy_client,
        collection="test_collection",
        operations=[operation],
        agents_host="http://dummy-agent",
    )

    with pytest.raises(ValueError) as exc_info:
        agent.update_all()

    assert "Append operations must use AppendPropertyOperation type" in str(
        exc_info.value
    )



from unittest.mock import Mock
from uuid import UUID, uuid4

import httpx
import pytest
import weaviate.classes as wvc

from weaviate_agents.personalization.personalization_agent import (
    PersonalizationAgent,
    PersonalizationAgentGetObjectsResponse,
    PersonalizedQuery,
)


class DummyClient:
    """A dummy client to simulate the Weaviate client for testing purposes."""

    def __init__(self):
        self._connection = self  # Simulate connection
        self.url = "http://test-url:443"  # Add url attribute
        self.additional_headers = {"Authorization": "test-token"}

    def get_current_bearer_token(self) -> str:
        """Simulate bearer token retrieval from connection.

        Returns:
            str: The test bearer token prefixed with 'Bearer '.
        """
        return "Bearer test-token"


@pytest.fixture
def agent() -> PersonalizationAgent:
    return PersonalizationAgent(
        client=DummyClient(),  # type: ignore
        reference_collection="TestCollection",
    )


def test_query(agent: PersonalizationAgent):
    query = agent.query(persona_id=uuid4())
    assert isinstance(query, PersonalizedQuery)


@pytest.mark.parametrize(
    "filters, filter_request",
    [
        (None, None),
        (
            wvc.query.Filter.by_property("title").not_equal("Avatar"),
            {"operator": "NotEqual", "target": "title", "value": "Avatar"},
        ),
        (
            (
                wvc.query.Filter.by_property("title").not_equal("Avatar")
                & wvc.query.Filter.by_property("genres").not_equal("comedy")
            ),
            {
                "combine": "and",
                "filters": [
                    {"operator": "NotEqual", "target": "title", "value": "Avatar"},
                    {"operator": "NotEqual", "target": "genres", "value": "comedy"},
                ],
            },
        ),
    ],
)
def test_get_objects(agent: PersonalizationAgent, monkeypatch, filters, filter_request):
    mock_response = {
        "objects": [
            {
                "uuid": "ae0fc31d-2f11-4e3a-a0f8-6e64795d7db9",
                "original_rank": 0,
                "personalized_rank": None,
                "properties": {
                    "title": "The Crush",
                    "genres": ["Drama", "Thriller", "Horror"],
                },
            },
            {
                "uuid": "f557e407-74db-453a-95b4-9357045be38d",
                "original_rank": 1,
                "personalized_rank": None,
                "properties": {
                    "title": "Armed Response",
                    "genres": ["Action", "Horror", "Thriller"],
                },
            },
        ],
        "ranking_rationale": None,
        "usage": {
            "requests": 0,
            "request_tokens": 0,
            "response_tokens": 0,
            "total_tokens": 0,
            "details": None,
        },
    }

    mock_httpx = Mock()
    mock_httpx.post.return_value = httpx.Response(
        status_code=200,
        request=httpx.Request("post", "http://dummy-agent"),
        json=mock_response,
    )
    monkeypatch.setattr(
        "weaviate_agents.personalization.personalization_agent.httpx", mock_httpx
    )

    persona_id = UUID("32a31598-42a8-4ed2-92d0-1273c7dbdcb0")
    recommendations = agent.get_objects(
        persona_id=persona_id,
        filters=filters,
    )
    expected_request = {
        "objects_request": {
            "persona_id": "32a31598-42a8-4ed2-92d0-1273c7dbdcb0",
            "limit": 10,
            "recent_interactions_count": 100,
            "exclude_interacted_items": True,
            "decay_rate": 0.1,
            "exclude_items": [],
            "use_agent_ranking": True,
            "explain_results": True,
            "instruction": None,
            "filters": filter_request,
        },
        "personalization_request": {
            "collection_name": "TestCollection",
            "headers": {"Authorization": "test-token"},
            "item_collection_vector_name": None,
            "create": False,
        },
    }

    mock_httpx.post.assert_called_once()
    request_json = mock_httpx.post.call_args.kwargs["json"]
    assert request_json == expected_request

    assert isinstance(recommendations, PersonalizationAgentGetObjectsResponse)
    assert recommendations.model_dump(mode="json") == mock_response



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "weaviate-agents-python-client"
copyright = "2025, Weaviate"
author = "Weaviate"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.autodoc_pydantic",
]

# Autodoc settings for pydantic
autodoc_pydantic_model_show_json = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_model_show_validator_members = False
autodoc_pydantic_model_show_field_summary = False
autodoc_pydantic_model_undoc_members = False
autodoc_pydantic_model_members = False

autodoc_typehints = "description"
autodoc_member_order = "bysource"
autodoc_dataclass_fields = False

# Make sure the target is unique
autosectionlabel_prefix_document = True

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "README.rst"]

suppress_warnings = [
    "docutils",
    "autodoc",
    "autosectionlabel",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


import re


def convert_markdown_links(lines):
    """Convert Markdown-style [text](url) links to reST-style `text <url>`_ links."""
    md_link_pattern = re.compile(r"\[([^\]]+)\]\((http[^\)]+)\)")
    return [md_link_pattern.sub(r"`\1 <\2>`_", line) for line in lines]


def autodoc_process_docstring(app, what, name, obj, options, lines):
    """Apply the conversion to all docstrings."""
    lines[:] = convert_markdown_links(lines)


def setup(app):
    app.add_css_file("custom.css")
    app.connect("autodoc-process-docstring", autodoc_process_docstring)



from typing import Optional
from uuid import UUID

import httpx
from weaviate.classes.config import DataType
from weaviate.client import WeaviateClient
from weaviate.collections.classes.filters import _Filters

from weaviate_agents.base import _BaseAgent
from weaviate_agents.personalization.classes import (
    GetObjectsRequest,
    Persona,
    PersonaInteraction,
    PersonaInteractionResponse,
    PersonalizationAgentGetObjectsResponse,
    PersonalizationRequest,
)
from weaviate_agents.personalization.query import PersonalizedQuery


class PersonalizationAgent(_BaseAgent[WeaviateClient]):
    """An agent for personalizing search results and queries based on persona interactions.

    Warning:
        Weaviate Agents - Personalization Agent is an early stage alpha product. The API is subject to
        breaking changes. Please ensure you are using the latest version of the client.

        For more information, see the [Weaviate Agents - Personalization Agent Docs](https://weaviate.io/developers/agents/personalization)
    """

    def __init__(
        self,
        client: WeaviateClient,
        reference_collection: str,
        agents_host: Optional[str] = None,
        vector_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ):
        super().__init__(client, agents_host)
        self._reference_collection = reference_collection
        self._vector_name = vector_name
        self._route = "/personalization"
        self._timeout = timeout

    @classmethod
    def create(
        cls,
        client: WeaviateClient,
        reference_collection: str,
        user_properties: Optional[dict[str, DataType]] = None,
        agents_host: Optional[str] = None,
        vector_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> "PersonalizationAgent":
        """Create a new Personalization Agent for a collection.

        Args:
            client: The Weaviate client
            reference_collection: The name of the collection to personalize
            user_properties: Optional dictionary of user properties and their data types
            agents_host: Optional host URL for the agents service
            vector_name: Optional name of the vector field to use
            timeout: Optional timeout for the request

        Returns:
            A new instance of the Personalization Agent
        """
        agent = cls(
            client=client,
            reference_collection=reference_collection,
            agents_host=agents_host,
            vector_name=vector_name,
        )
        agent._initialize(
            reference_collection,
            create=True,
            user_properties=user_properties,
            vector_name=vector_name,
            timeout=timeout,
        )
        return agent

    @classmethod
    def connect(
        cls,
        client: WeaviateClient,
        reference_collection: str,
        agents_host: Optional[str] = None,
        vector_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> "PersonalizationAgent":
        """Connect to an existing Personalization Agent for a collection.

        Args:
            client: The Weaviate client
            reference_collection: The name of the collection to connect to
            agents_host: Optional host URL for the agents service
            vector_name: Optional name of the vector field to use
            timeout: Optional timeout for the request

        Returns:
            An instance of the Personalization Agent
        """
        agent = cls(
            client=client,
            reference_collection=reference_collection,
            agents_host=agents_host,
            vector_name=vector_name,
        )
        agent._initialize(
            reference_collection, create=False, vector_name=vector_name, timeout=timeout
        )
        return agent

    def _initialize(
        self,
        reference_collection: str,
        create: bool = False,
        user_properties: Optional[dict[str, DataType]] = None,
        vector_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ):
        """Initialize the agent with the given reference collection and user properties.

        Args:
            reference_collection: The name of the collection to personalize
            create: Whether to create a new personalization agent
            user_properties: Optional dictionary of user properties and their data types
            vector_name: Optional name of the vector field to use
            timeout: Optional timeout for the request
        """
        request_data = {
            "collection_name": reference_collection,
            "headers": self._connection.additional_headers,
            "persona_properties": user_properties or {},
            "item_collection_vector_name": vector_name,
            "create": create,
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/",
            headers=self._headers,
            json=request_data,
            timeout=timeout,
        )

        if response.is_error:
            raise Exception(
                f"Failed to initialize personalization agent: {response.text}"
            )

    def add_persona(self, persona: Persona) -> None:
        """Add a persona to the Personalization Agent's persona collection.

        Args:
            persona: The persona to add. The persona must have a persona_id and properties that match the user properties
            defined when the Personalization Agent was created.
        """
        request_data = {
            "persona": persona.model_dump(mode="json"),
            "personalization_request": {
                "collection_name": self._reference_collection,
                "headers": self._connection.additional_headers,
                "item_collection_vector_name": self._vector_name,
                "create": False,
            },
        }
        response = httpx.post(
            f"{self._agents_host}{self._route}/persona",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to add persona: {response.text}")

    def update_persona(self, persona: Persona) -> None:
        """Update an existing persona in the Personalization Agent's persona collection.

        Args:
            persona: The persona to update. The persona must have a persona_id and properties that match
                    the user properties defined when the Personalization Agent was created.
        """
        request_data = {
            "persona": persona.model_dump(mode="json"),
            "personalization_request": {
                "collection_name": self._reference_collection,
                "headers": self._connection.additional_headers,
                "item_collection_vector_name": self._vector_name,
                "create": False,
            },
        }
        response = httpx.put(
            f"{self._agents_host}{self._route}/persona",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )
        if response.is_error:
            raise Exception(f"Failed to update persona: {response.text}")

    def get_persona(self, persona_id: UUID) -> Persona:
        """Get a persona by persona_id from the Personalization Agent's persona collection.

        Args:
            persona_id: The ID of the persona to retrieve

        Returns:
            The retrieved persona
        """
        request_data = {
            "collection_name": self._reference_collection,
            "headers": self._connection.additional_headers,
            "item_collection_vector_name": self._vector_name,
            "create": False,
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/persona/{str(persona_id)}",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to get persona: {response.text}")

        return Persona(**response.json())

    def delete_persona(self, persona_id: UUID) -> None:
        """Delete a persona by persona_id from the Personalization Agent's persona collection.

        Args:
            persona_id: The ID of the persona to delete
        """
        request_data = {
            "collection_name": self._reference_collection,
            "headers": self._connection.additional_headers,
            "item_collection_vector_name": self._vector_name,
            "create": False,
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/persona/delete/{str(persona_id)}",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to delete persona: {response.text}")

    def has_persona(self, persona_id: UUID) -> bool:
        """Check if a persona exists in the Personalization Agent's persona collection.

        Args:
            persona_id: The ID of the persona to check

        Returns:
            True if the persona exists, False otherwise
        """
        request_data = {
            "collection_name": self._reference_collection,
            "headers": self._connection.additional_headers,
            "item_collection_vector_name": self._vector_name,
            "create": False,
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/persona/{str(persona_id)}/exists",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to check persona existence: {response.text}")

        return response.json()["exists"]

    def add_interactions(
        self,
        interactions: list[PersonaInteraction],
        create_persona_if_not_exists: bool = True,
        remove_previous_interactions: bool = False,
    ) -> None:
        """Add interactions for personas to the Personalization Agent.

        Args:
            interactions: List of interactions to add. Each interaction can specify
                        `replace_previous_interactions=True` to replace that specific
                        item's interaction history.
            create_persona_if_not_exists: Whether to create personas that don't exist yet
            remove_previous_interactions: Whether to remove previous interactions for all items
                                       in the current batch. Setting this to True is equivalent
                                       to setting `replace_previous_interactions=True` for every
                                       interaction in the batch. Use with caution as it affects
                                       all items in the current batch.
        """
        request_data = {
            "interactions_request": {
                "interactions": [
                    interaction.model_dump(mode="json") for interaction in interactions
                ],
                "create_persona_if_not_exists": create_persona_if_not_exists,
                "remove_previous_interactions": remove_previous_interactions,
            },
            "personalization_request": {
                "collection_name": self._reference_collection,
                "headers": self._connection.additional_headers,
                "item_collection_vector_name": self._vector_name,
                "create": False,
            },
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/interactions",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to add interactions: {response.text}")

    def get_interactions(
        self, persona_id: UUID, interaction_type: str
    ) -> list[PersonaInteractionResponse]:
        """Get interactions for a specific persona filtered by interaction type.

        Args:
            persona_id: The ID of the persona to get interactions for
            interaction_type: The type of interaction to filter by (e.g. "positive", "negative")

        Returns:
            List of matching interactions for the persona
        """
        request_data = {
            "interaction_request": {
                "persona_id": str(persona_id),
                "interaction_type": interaction_type,
            },
            "personalization_request": {
                "collection_name": self._reference_collection,
                "headers": self._connection.additional_headers,
                "item_collection_vector_name": self._vector_name,
                "create": False,
            },
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/interactions/get",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to get interactions: {response.text}")

        return [
            PersonaInteractionResponse(**interaction) for interaction in response.json()
        ]

    def get_objects(
        self,
        persona_id: UUID,
        limit: int = 10,
        recent_interactions_count: int = 100,
        exclude_interacted_items: bool = True,
        decay_rate: float = 0.1,
        exclude_items: list[str] = [],
        use_agent_ranking: bool = True,
        explain_results: bool = True,
        instruction: Optional[str] = None,
        filters: Optional[_Filters] = None,
    ) -> PersonalizationAgentGetObjectsResponse:
        """Get Personalized objects for a specific persona.

        Args:
            persona_id: The ID of the persona to get objects for
            limit: The maximum number of objects to return
            recent_interactions_count: The number of recent interactions to consider
            exclude_interacted_items: Whether to exclude items that have been interacted with
            decay_rate: The decay rate for the personalization algorithm
            exclude_items: List of items to exclude from the results
            use_agent_ranking: Whether to use agent ranking for the results
            explain_results: Whether to explain the results
            instruction: Optional instruction to guide the personalization process
            filters: Optional filters to apply to the results
        """
        objects_request = GetObjectsRequest(
            persona_id=persona_id,
            limit=limit,
            recent_interactions_count=recent_interactions_count,
            exclude_interacted_items=exclude_interacted_items,
            decay_rate=decay_rate,
            exclude_items=exclude_items,
            use_agent_ranking=use_agent_ranking,
            explain_results=explain_results,
            instruction=instruction,
            filters=filters,
        )
        request_data = {
            "objects_request": objects_request.model_dump(mode="json"),
            "personalization_request": {
                "collection_name": self._reference_collection,
                "headers": self._connection.additional_headers,
                "item_collection_vector_name": self._vector_name,
                "create": False,
            },
        }

        response = httpx.post(
            f"{self._agents_host}{self._route}/objects",
            headers=self._headers,
            json=request_data,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(f"Failed to get objects: {response.text}")

        return PersonalizationAgentGetObjectsResponse(**response.json())

    @classmethod
    def exists(
        cls,
        client: WeaviateClient,
        reference_collection: str,
        agents_host: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> bool:
        """Check if a persona collection exists for a given reference collection.

        Args:
            client: The Weaviate client
            reference_collection: The name of the collection to check
            agents_host: Optional host URL for the agents service
            timeout: Optional timeout for the request

        Returns:
            True if the persona collection exists, False otherwise
        """
        # Initialize base values from client
        base_agent = cls(client, reference_collection, agents_host=agents_host)

        response = httpx.get(
            f"{base_agent._agents_host}{base_agent._route}/exists/{reference_collection}",
            headers=base_agent._headers,
            timeout=timeout,
        )

        if response.is_error:
            raise Exception(
                f"Failed to check if persona collection exists: {response.text}"
            )

        return response.json()["persona_collection_exists"]

    def query(
        self,
        persona_id: UUID,
        strength: float = 0.5,
        overfetch_factor: float = 1.5,
        recent_interactions_count: int = 100,
        decay_rate: float = 0.1,
    ) -> PersonalizedQuery:
        personalization_request = PersonalizationRequest(
            collection_name=self._reference_collection,
            headers=self._connection.additional_headers,
            item_collection_vector_name=self._vector_name,
            create=False,
        )
        return PersonalizedQuery(
            agents_host=self._agents_host,
            headers=self._headers,
            persona_id=persona_id,
            personalization_request=personalization_request,
            timeout=self._timeout,
            strength=strength,
            overfetch_factor=overfetch_factor,
            recent_interactions_count=recent_interactions_count,
            decay_rate=decay_rate,
        )



from __future__ import annotations

from typing import Annotated, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, PlainSerializer
from typing_extensions import TypedDict
from weaviate.classes.query import Move, Rerank
from weaviate.collections.classes.filters import (
    _FilterAnd,
    _FilterOr,
    _Filters,
    _FilterValue,
)
from weaviate.collections.classes.grpc import (
    METADATA,
    HybridFusion,
    HybridVectorType,
    NearVectorInputType,
    TargetVectorJoinType,
    _HybridNearText,
    _HybridNearVector,
)
from weaviate.collections.classes.internal import ReturnProperties, ReturnReferences


class NearTextQueryParameters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    query_method: Literal["near_text"] = "near_text"

    query: Union[List[str], str]
    certainty: Union[int, float, None] = None
    distance: Union[int, float, None] = None
    move_to: Optional[Annotated[Move, serialise_move]] = None
    move_away: Optional[Annotated[Move, serialise_move]] = None
    limit: Union[int, None] = None
    offset: Union[int, None] = None
    auto_limit: Union[int, None] = None
    filters: Optional[Annotated[_Filters, serialise_filter]] = None
    # group_by: Optional[weaviate.collections.classes.grpc.GroupBy] = None
    rerank: Optional[Rerank] = None
    target_vector: Optional[TargetVectorJoinType] = None
    include_vector: Union[bool, str, List[str]] = False
    return_metadata: Optional[METADATA] = None
    return_properties: Optional[ReturnProperties[dict]] = None
    return_references: Optional[ReturnReferences[dict]] = None


class BM25QueryParameters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    query_method: Literal["bm25"] = "bm25"

    query: Union[str, None]
    query_properties: Union[List[str], None] = None
    limit: Union[int, None] = None
    offset: Union[int, None] = None
    auto_limit: Union[int, None] = None
    filters: Optional[Annotated[_Filters, serialise_filter]] = None
    # group_by: Optional[weaviate.collections.classes.grpc.GroupBy] = None,
    rerank: Optional[Rerank] = None
    include_vector: Union[bool, str, List[str]] = False
    return_metadata: Optional[METADATA] = None
    return_properties: Optional[ReturnProperties[dict]] = None
    return_references: Optional[ReturnReferences[dict]] = None


class HybridQueryParameters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    query_method: Literal["hybrid"] = "hybrid"

    query: Union[str, None]
    alpha: Union[int, float] = 0.7
    vector: Union[Annotated[HybridVectorType, serialise_hybrid_vector_type], None] = (
        None
    )
    query_properties: Union[List[str], None] = None
    fusion_type: Optional[HybridFusion] = None
    max_vector_distance: Union[int, float, None] = None
    limit: Union[int, None] = None
    offset: Union[int, None] = None
    auto_limit: Union[int, None] = None
    filters: Optional[Annotated[_Filters, serialise_filter]] = None
    # group_by: Optional[weaviate.collections.classes.grpc.GroupBy] = None,
    rerank: Optional[Rerank] = None
    target_vector: Optional[TargetVectorJoinType] = None
    include_vector: Union[bool, str, List[str]] = False
    return_metadata: Optional[METADATA] = None
    return_properties: Optional[ReturnProperties[dict]] = None
    return_references: Optional[ReturnReferences[dict]] = None


QueryParameters = Union[
    NearTextQueryParameters, BM25QueryParameters, HybridQueryParameters
]


class QueryRequest(BaseModel):
    persona_id: UUID
    strength: float
    recent_interactions_count: int
    decay_rate: float
    overfetch_factor: float
    query_parameters: QueryParameters = Field(discriminator="query_method")


class _MoveSerialise(TypedDict):
    force: float
    objects: Optional[List[str]]
    concepts: Optional[List[str]]


@PlainSerializer
def serialise_move(move: Optional[Move]) -> Optional[_MoveSerialise]:
    if move is None:
        return None
    return _MoveSerialise(
        force=move.force, objects=move._objects_list, concepts=move._concepts_list
    )


class _FilterAndOrSerialise(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    combine: Literal["and", "or"]
    filters: list[_Filters]


@PlainSerializer
def serialise_filter(
    filter_and_or_value: Union[_FilterAnd, _FilterOr, _FilterValue],
) -> Union[_FilterValue, _FilterAndOrSerialise]:
    if isinstance(filter_and_or_value, _FilterValue):
        return filter_and_or_value

    if isinstance(filter_and_or_value, _FilterAnd):
        combine: Literal["and", "or"] = "and"
    elif isinstance(filter_and_or_value, _FilterOr):
        combine = "or"
    else:
        raise TypeError(f"Unknown filter type {type(filter_and_or_value)}")
    return _FilterAndOrSerialise(combine=combine, filters=filter_and_or_value.filters)


# A set of models to help serialise HybridVectorType, which is a union of
# NearVectorInputType (itself a union over std types), _HybridNearText and _HybridNearVector.
class _HybridNearTextSerialise(_HybridNearText):
    serialised_class: Literal["_HybridNearText"] = "_HybridNearText"

    move_to: Annotated[Optional[Move], serialise_move] = None
    move_away: Annotated[Optional[Move], serialise_move] = None


def _serialise_hybrid_near_text(model: _HybridNearText) -> _HybridNearTextSerialise:
    return _HybridNearTextSerialise.model_validate(model.model_dump())


class _HybridNearVectorSerialise(BaseModel):
    serialised_class: Literal["_HybridNearVector"] = "_HybridNearVector"

    vector: NearVectorInputType
    distance: Optional[float]
    certainty: Optional[float]


def _serialise_hybrid_near_vector(
    model: _HybridNearVector,
) -> _HybridNearVectorSerialise:
    return _HybridNearVectorSerialise(
        vector=model.vector,
        distance=model.distance,
        certainty=model.certainty,
    )


@PlainSerializer
def serialise_hybrid_vector_type(
    vector: HybridVectorType,
) -> Union[_HybridNearTextSerialise, _HybridNearVectorSerialise, NearVectorInputType]:
    if isinstance(vector, _HybridNearText):
        return _serialise_hybrid_near_text(vector)
    elif isinstance(vector, _HybridNearVector):
        return _serialise_hybrid_near_vector(vector)
    else:
        return vector



from ...classes.core import Usage
from .collection import QueryAgentCollectionConfig
from .response import (
    AggregationResult,
    AggregationResultWithCollection,
    BooleanArrayPropertyFilter,
    BooleanMetrics,
    BooleanPropertyAggregation,
    BooleanPropertyFilter,
    ComparisonOperator,
    DateArrayPropertyFilter,
    DateMetrics,
    DatePropertyAggregation,
    DatePropertyFilter,
    GeoPropertyFilter,
    IntegerArrayPropertyFilter,
    IntegerPropertyAggregation,
    IntegerPropertyFilter,
    NumericMetrics,
    QueryAgentResponse,
    QueryResult,
    QueryResultWithCollection,
    Source,
    TextArrayPropertyFilter,
    TextMetrics,
    TextPropertyAggregation,
    TextPropertyFilter,
    UnknownPropertyAggregation,
    UnknownPropertyFilter,
)

__all__ = [
    "QueryAgentCollectionConfig",
    "QueryAgentResponse",
    "Source",
    "ComparisonOperator",
    "IntegerPropertyFilter",
    "TextPropertyFilter",
    "BooleanPropertyFilter",
    "QueryResult",
    "NumericMetrics",
    "TextMetrics",
    "BooleanMetrics",
    "IntegerPropertyAggregation",
    "TextPropertyAggregation",
    "BooleanPropertyAggregation",
    "AggregationResult",
    "Usage",
    "AggregationResultWithCollection",
    "QueryResultWithCollection",
    "BooleanArrayPropertyFilter",
    "DateArrayPropertyFilter",
    "DateMetrics",
    "DatePropertyAggregation",
    "DatePropertyFilter",
    "IntegerArrayPropertyFilter",
    "TextArrayPropertyFilter",
    "GeoPropertyFilter",
    "UnknownPropertyAggregation",
    "UnknownPropertyFilter",
]



from typing import Union

from pydantic import BaseModel


class QueryAgentCollectionConfig(BaseModel):
    """A collection configuration for the QueryAgent.

    Attributes:
        name: The name of the collection to query.
        tenant: Tenant name for collections with multi-tenancy enabled.
        view_properties: Optional list of property names the agent has the ability to view
            for this specific collection.
        target_vector: Optional target vector name(s) for collections with named vectors.
            Can be a single vector name or a list of vector names.
    """

    name: str
    tenant: Union[str, None] = None
    view_properties: Union[list[str], None] = None
    target_vector: Union[str, list[str], None] = None



from typing import Annotated, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict
from weaviate.collections.classes.filters import _Filters

from weaviate_agents.personalization.classes.query import serialise_filter


class PersonalizationRequest(BaseModel):
    collection_name: str
    create: bool = True
    headers: Union[Dict[str, str], None] = None
    persona_properties: Union[Dict[str, str], None] = None
    item_collection_vector_name: Union[str, None] = None


class GetObjectsRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    persona_id: UUID
    limit: int = 10
    recent_interactions_count: int = 100
    exclude_interacted_items: bool = True
    decay_rate: float = 0.1
    exclude_items: List[str] = []
    use_agent_ranking: bool = True
    explain_results: bool = True
    instruction: Optional[str] = None
    filters: Optional[Annotated[_Filters, serialise_filter]] = None



from typing import List, Union
from uuid import UUID

import httpx
from weaviate.client import WeaviateClient

from weaviate_agents.base import _BaseAgent
from weaviate_agents.transformation.classes import (
    AppendPropertyOperation,
    OperationStep,
    OperationType,
    TransformationResponse,
    UpdatePropertyOperation,
)


class TransformationAgent(_BaseAgent[WeaviateClient]):
    """An agent for running large scale transformations on data in Weaviate.

    Warning:
        Weaviate Agents - Transformation Agent is an early stage alpha product. The API is subject to
        breaking changes. Please ensure you are using the latest version of the client.

        For more information, see the [Weaviate Agents - Transformation Agent Docs](https://weaviate.io/developers/agents/transformation)
    """

    def __init__(
        self,
        client: WeaviateClient,
        collection: str,
        operations: List[OperationStep],
        agents_host: Union[str, None] = None,
        timeout: Union[int, None] = None,
    ):
        """Initialize the TransformationAgent.

        Args:
            client: The Weaviate client connected to a Weaviate Cloud cluster.
            collection: The collection to perform transformations on.
            operations: A list of operations to execute on the collection.
            agents_host: Optional host of the agents service.
            timeout: The timeout for the request. Defaults to 60 seconds.
        """
        super().__init__(
            client=client,
            agents_host=agents_host,
        )
        self.collection = collection
        self.operations = operations

        self._timeout = 60 if timeout is None else timeout

        self.t_host = f"{self._agents_host}/transformation"

    def _build_request_operations(self) -> list:
        request_operations = []
        for operation in self.operations:
            if operation.operation_type == OperationType.APPEND:
                if not isinstance(operation, AppendPropertyOperation):
                    raise ValueError(
                        "Append operations must use AppendPropertyOperation type"
                    )
                request_operation = {
                    "type": "create",
                    "instruction": operation.instruction,
                    "view_properties": operation.view_properties,
                    "on_properties": [
                        {
                            "name": operation.property_name,
                            "data_type": operation.data_type.value,
                        }
                    ],
                }
            elif operation.operation_type == OperationType.UPDATE:
                if not isinstance(operation, UpdatePropertyOperation):
                    raise ValueError(
                        "Update operations must use UpdatePropertyOperation type"
                    )
                request_operation = {
                    "type": "update",
                    "instruction": operation.instruction,
                    "view_properties": operation.view_properties,
                    "on_properties": [operation.property_name],
                }
            else:
                raise ValueError(
                    f"Unsupported operation type: {operation.operation_type}. "
                    "Only APPEND and UPDATE operations are supported."
                )
            request_operations.append(request_operation)
        return request_operations

    def _post_transformation(self, request: dict) -> TransformationResponse:
        with httpx.Client(timeout=self._timeout) as client:
            response = client.post(
                self.t_host + "/properties",
                json=request,
                headers=self._headers,
            )
            if response.is_error:
                raise Exception(response.text)
            return TransformationResponse(**response.json())

    def update_all(self) -> TransformationResponse:
        """Triggers all configured transformation operations on the collection.

        Returns:
            TransformationResponse: response with workflow ID for tracking transformation progress.

        Raises:
            httpx.HTTPError: If there is an error communicating with the transformation service.
            ValueError: If the operations are not properly configured or if there are duplicate
                property operations.
        """
        request = {
            "collection": self.collection,
            "operations": self._build_request_operations(),
            "headers": self._connection.additional_headers,
        }
        return self._post_transformation(request)

    def get_status(self, workflow_id: str) -> dict:
        """Check the status of a transformation workflow.

        Args:
            workflow_id: The ID of the workflow to check, obtained from TransformationResponse

        Returns:
            The status response from the transformation service

        Raises:
            httpx.HTTPError: If there is an error communicating with the transformation service
        """
        with httpx.Client(timeout=self._timeout) as client:
            response = client.get(
                f"{self.t_host}/properties/status/{workflow_id}",
                headers=self._headers,
            )

            if response.is_error:
                raise Exception(response.text)

            return response.json()

    def update_by_uuids(self, uuids: List[UUID]) -> TransformationResponse:
        """Triggers all configured transformation operations on the specified UUIDs in the collection.

        Args:
            uuids: List of UUIDs to process.

        Returns:
            TransformationResponse: response with workflow ID for tracking transformation progress.

        Raises:
            httpx.HTTPError: If there is an error communicating with the transformation service.
            ValueError: If the operations are not properly configured or if there are duplicate
                property operations.
        """
        request = {
            "collection": self.collection,
            "operations": self._build_request_operations(),
            "headers": self._connection.additional_headers,
            "uuids": [str(u) for u in uuids],
        }
        return self._post_transformation(request)



import httpx
import pytest
from pydantic import ValidationError

from weaviate_agents.classes.query import QueryAgentCollectionConfig, QueryAgentResponse
from weaviate_agents.query import AsyncQueryAgent, QueryAgent


class DummyClient:
    """A dummy client to simulate the Weaviate client for testing purposes."""

    def __init__(self):
        self._connection = self  # Simulate connection
        self.url = "http://test-url:443"  # Add url attribute
        self.additional_headers = {"Authorization": "test-token"}

    def get_current_bearer_token(self) -> str:
        """Simulate bearer token retrieval from connection.

        Returns:
            str: The test bearer token prefixed with 'Bearer '.
        """
        return "Bearer test-token"


class FakeResponse:
    """A fake HTTP response for testing purposes.

    Attributes:
        status_code (int): The HTTP status code.
        _json (dict): The JSON data to return.
    """

    def __init__(self, status_code: int, json_data: dict):
        self.status_code = status_code
        self.is_error = 400 <= status_code <= 599
        self._json = json_data
        self.text = json_data

    def json(self) -> dict:
        """Return the JSON data of the fake response.

        Returns:
            dict: The JSON data.
        """
        return self._json


def fake_post_success(*args, **kwargs) -> FakeResponse:
    """Simulate a successful HTTP POST response.

    Returns:
        FakeResponse: A fake HTTP response with status code 200.
    """
    json_data = {
        "original_query": "test query",
        "collection_names": ["test_collection"],
        "searches": [
            [
                {
                    "collection": "test_collection",
                    "queries": ["Test query!"],
                    "filters": [
                        [
                            {
                                "filter_type": "integer",
                                "property_name": "prop_int",
                                "operator": "=",
                                "value": 1.0,
                            },
                            {
                                "filter_type": "integer_array",
                                "property_name": "prop_int_aray",
                                "operator": "contains_all",
                                "value": [1.0, 2.0],
                            },
                            {
                                "filter_type": "text",
                                "property_name": "prop_text",
                                "operator": "LIKE",
                                "value": "*something*",
                            },
                            {
                                "filter_type": "text_array",
                                "property_name": "prop_text_array",
                                "operator": "contains_any",
                                "value": ["one", "two"],
                            },
                            {
                                "filter_type": "boolean",
                                "property_name": "prop_bool",
                                "operator": "=",
                                "value": True,
                            },
                            {
                                "filter_type": "boolean_array",
                                "property_name": "prop_bool_array",
                                "operator": "contains_any",
                                "value": [True, False],
                            },
                            {
                                "filter_type": "date_range",
                                "property_name": "prop_date",
                                "value": {
                                    "date_from": "2025-01-01T12:01:23Z",
                                    "date_to": "2025-01-02T12:01:23Z",
                                    "inclusive_from": True,
                                    "inclusive_to": True,
                                },
                            },
                            {
                                "filter_type": "date_range",
                                "property_name": "prop_date",
                                "value": {
                                    "exact_timestamp": "2025-01-01T12:01:23Z",
                                    "operator": "=",
                                },
                            },
                            {
                                "filter_type": "date_array",
                                "property_name": "prop_date_array",
                                "operator": "contains_all",
                                "value": [
                                    "2025-01-01T12:01:23Z",
                                    "2025-01-02T12:01:23Z",
                                ],
                            },
                            {
                                "filter_type": "geo",
                                "property_name": "prop_geo",
                                "latitude": 10.0,
                                "longitude": 20.0,
                                "max_distance_meters": 30.0,
                            },
                            {
                                "filter_type": "something_new",
                                "property_name": "strange_property",
                                "value": "xyz",
                            },
                        ]
                    ],
                    "filter_operators": "AND",
                }
            ]
        ],
        "aggregations": [
            [
                {
                    "collection": "test_collection",
                    "search_query": None,
                    "groupby_property": None,
                    "aggregations": [
                        {
                            "aggregation_type": "integer",
                            "property_name": "prop_int",
                            "metrics": "MEAN",
                        },
                        {
                            "aggregation_type": "text",
                            "property_name": "prop_text",
                            "metrics": "COUNT",
                            "top_occurrences_limit": 10,
                        },
                        {
                            "aggregation_type": "boolean",
                            "property_name": "prop_bool",
                            "metrics": "PERCENTAGE_TRUE",
                        },
                        {
                            "aggregation_type": "date",
                            "property_name": "prop_date",
                            "metrics": "MAXIMUM",
                        },
                        {
                            "aggregation_type": "something_new",
                            "property_name": "strange_property",
                            "metrics": "XYZ",
                        },
                    ],
                    "filters": [],
                }
            ]
        ],
        "sources": [{"object_id": "123", "collection": "test_collection"}],
        "usage": {
            "requests": 1,
            "request_tokens": 10,
            "response_tokens": 20,
            "total_tokens": 30,
            "details": {},
        },
        "total_time": 0.1,
        "aggregation_answer": None,
        "has_aggregation_answer": False,
        "has_search_answer": False,
        "is_partial_answer": False,
        "missing_information": [],
        "final_answer": "final answer",
    }
    return FakeResponse(200, json_data)


async def fake_async_post_success(*args, **kwargs) -> FakeResponse:
    return fake_post_success(*args, **kwargs)


def fake_post_failure(*args, **kwargs) -> FakeResponse:
    """Simulate a failed HTTP POST response.

    Returns:
        FakeResponse: A fake HTTP response with a non-200 status code.
    """
    json_data = {
        "error": {
            "message": "Test error message",
            "code": "test_error_code",
            "details": {"info": "test detail"},
        }
    }
    return FakeResponse(400, json_data)


async def fake_async_post_failure(*args, **kwargs):
    return fake_post_failure(*args, **kwargs)


def test_run_success(monkeypatch):
    """Test that QueryAgent.run returns a valid QueryAgentResponse when the HTTP call is successful.

    Returns:
        None.
    """
    monkeypatch.setattr(httpx, "post", fake_post_success)
    dummy_client = DummyClient()
    agent = QueryAgent(
        dummy_client, ["test_collection"], agents_host="http://dummy-agent"
    )
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = agent.run("test query")
    assert isinstance(result, QueryAgentResponse)
    assert result.original_query == "test query"
    assert result.collection_names == ["test_collection"]
    assert result.total_time == 0.1
    assert result.final_answer == "final answer"


async def test_async_run_success(monkeypatch):
    monkeypatch.setattr(httpx.AsyncClient, "post", fake_async_post_success)
    dummy_client = DummyClient()
    agent = AsyncQueryAgent(
        dummy_client, ["test_collection"], agents_host="http://dummy-agent"
    )
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = await agent.run("test query")
    assert isinstance(result, QueryAgentResponse)
    assert result.original_query == "test query"
    assert result.collection_names == ["test_collection"]
    assert result.total_time == 0.1
    assert result.final_answer == "final answer"


def test_run_failure(monkeypatch):
    """Test that QueryAgent.run raises an exception when the HTTP response indicates an error.

    Returns:
        None.
    """
    monkeypatch.setattr(httpx, "post", fake_post_failure)
    dummy_client = DummyClient()
    agent = QueryAgent(
        dummy_client, ["test_collection"], agents_host="http://dummy-agent"
    )
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    with pytest.raises(Exception) as exc_info:
        agent.run("failure query")

    assert (
        str(exc_info.value)
        == "{'error': {'message': 'Test error message', 'code': 'test_error_code', 'details': {'info': 'test detail'}}}"
    )


async def test_async_run_failure(monkeypatch):
    monkeypatch.setattr(httpx.AsyncClient, "post", fake_async_post_failure)
    dummy_client = DummyClient()
    agent = AsyncQueryAgent(
        dummy_client, ["test_collection"], agents_host="http://dummy-agent"
    )
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    with pytest.raises(Exception) as exc_info:
        await agent.run("failure query")

    assert (
        str(exc_info.value)
        == "{'error': {'message': 'Test error message', 'code': 'test_error_code', 'details': {'info': 'test detail'}}}"
    )


def test_query_agent_response_model_validation():
    """Test that the QueryAgentResponse model raises a ValidationError when required fields are missing.

    Returns:
        None.
    """
    incomplete_data = {"original_query": "incomplete query"}
    with pytest.raises(ValidationError):
        QueryAgentResponse(**incomplete_data)


def test_run_with_target_vector(monkeypatch):
    """Test that QueryAgent.run correctly passes the target_vector argument in the request body."""
    captured = {}

    def fake_post_with_capture(url, headers=None, json=None, timeout=None):
        captured["json"] = json
        # Return a normal successful response
        return fake_post_success()

    monkeypatch.setattr(httpx, "post", fake_post_with_capture)
    dummy_client = DummyClient()
    agent = QueryAgent(dummy_client, agents_host="http://dummy-agent")
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    # Test with single target vector
    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = agent.run(
            "test query",
            collections=[
                QueryAgentCollectionConfig(
                    name="test_collection", target_vector="my_vector"
                )
            ],
        )
    assert isinstance(result, QueryAgentResponse)
    assert captured["json"]["collections"][0]["target_vector"] == "my_vector"

    # Test with multiple target vectors
    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = agent.run(
            "test query",
            collections=[
                QueryAgentCollectionConfig(
                    name="test_collection",
                    target_vector=["first_vector", "second_vector"],
                )
            ],
        )
    assert isinstance(result, QueryAgentResponse)
    assert captured["json"]["collections"][0]["target_vector"] == [
        "first_vector",
        "second_vector",
    ]


async def test_async_run_with_target_vector(monkeypatch):
    captured = {}

    async def fake_async_post_with_capture(*args, **kwargs):
        captured["json"] = kwargs.get("json")
        return await fake_async_post_success()

    monkeypatch.setattr(httpx.AsyncClient, "post", fake_async_post_with_capture)
    dummy_client = DummyClient()
    agent = AsyncQueryAgent(dummy_client, agents_host="http://dummy-agent")
    agent._connection = dummy_client
    agent._headers = dummy_client.additional_headers

    # Test with single target vector
    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = await agent.run(
            "test query",
            collections=[
                QueryAgentCollectionConfig(
                    name="test_collection", target_vector="my_vector"
                )
            ],
        )
    assert isinstance(result, QueryAgentResponse)
    assert captured["json"]["collections"][0]["target_vector"] == "my_vector"

    # Test with multiple target vectors
    with pytest.warns(UserWarning):
        # Expect a warning when parsing the unkown "something_new" filter/aggregation
        result = await agent.run(
            "test query",
            collections=[
                QueryAgentCollectionConfig(
                    name="test_collection",
                    target_vector=["first_vector", "second_vector"],
                )
            ],
        )
    assert isinstance(result, QueryAgentResponse)
    assert captured["json"]["collections"][0]["target_vector"] == [
        "first_vector",
        "second_vector",
    ]



from typing import Dict, Union

from pydantic import BaseModel


class Usage(BaseModel):
    requests: Union[int, str] = 0
    request_tokens: Union[int, str, None] = None
    response_tokens: Union[int, str, None] = None
    total_tokens: Union[int, str, None] = None
    details: Union[Dict[str, int], Dict[str, str], None] = None



import warnings
from enum import Enum
from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from weaviate_agents.classes.core import Usage
from weaviate_agents.utils import print_query_agent_response


class ComparisonOperator(str, Enum):
    EQUALS = "="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_EQUAL = "<="
    GREATER_EQUAL = ">="
    NOT_EQUALS = "!="
    LIKE = "LIKE"
    CONTAINS_ANY = "contains_any"
    CONTAINS_ALL = "contains_all"


class KnownFilterType(str, Enum):
    INTEGER = "integer"
    INTEGER_ARRAY = "integer_array"
    TEXT = "text"
    TEXT_ARRAY = "text_array"
    BOOLEAN = "boolean"
    BOOLEAN_ARRAY = "boolean_array"
    DATE = "date_range"
    DATE_ARRAY = "date_array"
    GEO = "geo"


class KnownPropertyFilterBase(BaseModel):
    filter_type: KnownFilterType
    property_name: str


class IntegerPropertyFilter(KnownPropertyFilterBase):
    """Filter numeric properties using comparison operators."""

    filter_type: Literal[KnownFilterType.INTEGER] = Field(
        repr=False, default=KnownFilterType.INTEGER
    )

    operator: ComparisonOperator
    value: float


class IntegerArrayPropertyFilter(KnownPropertyFilterBase):
    """Filter numeric-array properties using comparison operators."""

    filter_type: Literal[KnownFilterType.INTEGER_ARRAY] = Field(
        repr=False, default=KnownFilterType.INTEGER_ARRAY
    )

    operator: ComparisonOperator
    value: list[float]


class TextPropertyFilter(KnownPropertyFilterBase):
    """Filter text properties using equality or LIKE operators."""

    filter_type: Literal[KnownFilterType.TEXT] = Field(
        repr=False, default=KnownFilterType.TEXT
    )

    operator: ComparisonOperator
    value: str


class TextArrayPropertyFilter(KnownPropertyFilterBase):
    """Filter text-array properties using equality or LIKE operators."""

    filter_type: Literal[KnownFilterType.TEXT_ARRAY] = Field(
        repr=False, default=KnownFilterType.TEXT_ARRAY
    )

    operator: ComparisonOperator
    value: list[str]


class BooleanPropertyFilter(KnownPropertyFilterBase):
    """Filter boolean properties using equality operators."""

    filter_type: Literal[KnownFilterType.BOOLEAN] = Field(
        repr=False, default=KnownFilterType.BOOLEAN
    )

    operator: ComparisonOperator
    value: bool


class BooleanArrayPropertyFilter(KnownPropertyFilterBase):
    """Filter boolean-array properties using equality operators."""

    filter_type: Literal[KnownFilterType.BOOLEAN_ARRAY] = Field(
        repr=False, default=KnownFilterType.BOOLEAN_ARRAY
    )

    operator: ComparisonOperator
    value: list[bool]


class DateExact(BaseModel):
    exact_timestamp: str
    operator: ComparisonOperator


class DateRange(BaseModel):
    date_from: Union[str, None]
    date_to: Union[str, None]
    inclusive_from: bool
    inclusive_to: bool


class DatePropertyFilter(KnownPropertyFilterBase):
    """Filter datetime properties using equality operators."""

    filter_type: Literal[KnownFilterType.DATE] = Field(
        repr=False, default=KnownFilterType.DATE
    )

    value: Union[DateExact, DateRange]


class DateArrayPropertyFilter(KnownPropertyFilterBase):
    """Filter datetime properties using equality operators."""

    filter_type: Literal[KnownFilterType.DATE_ARRAY] = Field(
        repr=False, default=KnownFilterType.DATE_ARRAY
    )

    operator: ComparisonOperator
    value: list[str]


class GeoPropertyFilter(KnownPropertyFilterBase):
    """Filter geo-coordinates properties."""

    filter_type: Literal[KnownFilterType.GEO] = Field(
        repr=False, default=KnownFilterType.GEO
    )

    latitude: float
    longitude: float
    max_distance_meters: float


class UnknownPropertyFilter(BaseModel):
    """Catch-all filter for unknown filter types, to preserve future back-compatibility."""

    model_config = ConfigDict(extra="allow")
    filter_type: None

    @field_validator("filter_type", mode="before")
    @classmethod
    def ensure_filter_type_unknown(cls, value: Any) -> None:
        if value in set(KnownFilterType):
            raise ValueError(
                f"{value} is an known filter type, but validation failed, "
                "so the response was not as expected. "
                "Try upgrading the weaviate-agents package to a new version."
            )
        return None

    def model_post_init(self, context: Any) -> None:
        warnings.warn(
            f"The filter_type {self.filter_type} wasn't recognised. "
            "Try upgrading the weaviate-agents package to a new version."
        )


PropertyFilter = Union[
    IntegerPropertyFilter,
    IntegerArrayPropertyFilter,
    TextPropertyFilter,
    TextArrayPropertyFilter,
    BooleanPropertyFilter,
    BooleanArrayPropertyFilter,
    DatePropertyFilter,
    DateArrayPropertyFilter,
    GeoPropertyFilter,
    UnknownPropertyFilter,
]


class QueryResult(BaseModel):
    queries: list[str]
    filters: list[list[PropertyFilter]] = []
    filter_operators: Literal["AND", "OR"]


class NumericMetrics(str, Enum):
    COUNT = "COUNT"
    MAX = "MAXIMUM"
    MEAN = "MEAN"
    MEDIAN = "MEDIAN"
    MIN = "MINIMUM"
    MODE = "MODE"
    SUM = "SUM"
    TYPE = "TYPE"


class TextMetrics(str, Enum):
    COUNT = "COUNT"
    TYPE = "TYPE"
    TOP_OCCURRENCES = "TOP_OCCURRENCES"


class BooleanMetrics(str, Enum):
    COUNT = "COUNT"
    TYPE = "TYPE"
    TOTAL_TRUE = "TOTAL_TRUE"
    TOTAL_FALSE = "TOTAL_FALSE"
    PERCENTAGE_TRUE = "PERCENTAGE_TRUE"
    PERCENTAGE_FALSE = "PERCENTAGE_FALSE"


class DateMetrics(str, Enum):
    COUNT = "COUNT"
    MAX = "MAXIMUM"
    MEDIAN = "MEDIAN"
    MIN = "MINIMUM"
    MODE = "MODE"


class KnownAggregationType(str, Enum):
    INTEGER = "integer"
    TEXT = "text"
    BOOLEAN = "boolean"
    DATE = "date"


class KnownPropertyAggregationBase(BaseModel):
    aggregation_type: KnownAggregationType
    property_name: str


class IntegerPropertyAggregation(KnownPropertyAggregationBase):
    """Aggregate numeric properties using statistical functions."""

    aggregation_type: Literal[KnownAggregationType.INTEGER] = Field(
        repr=False, default=KnownAggregationType.INTEGER
    )
    metrics: NumericMetrics


class TextPropertyAggregation(KnownPropertyAggregationBase):
    """Aggregate text properties using frequency analysis."""

    aggregation_type: Literal[KnownAggregationType.TEXT] = Field(
        repr=False, default=KnownAggregationType.TEXT
    )
    metrics: TextMetrics
    top_occurrences_limit: Optional[int] = None


class BooleanPropertyAggregation(KnownPropertyAggregationBase):
    """Aggregate boolean properties using statistical functions."""

    aggregation_type: Literal[KnownAggregationType.BOOLEAN] = Field(
        repr=False, default=KnownAggregationType.BOOLEAN
    )
    metrics: BooleanMetrics


class DatePropertyAggregation(KnownPropertyAggregationBase):
    """Aggregate datetime properties using statistical functions."""

    aggregation_type: Literal[KnownAggregationType.DATE] = Field(
        repr=False, default=KnownAggregationType.DATE
    )
    metrics: DateMetrics


class UnknownPropertyAggregation(BaseModel):
    """Catch-all aggregation for unknown aggregation types, to preserve future back-compatibility."""

    model_config = ConfigDict(extra="allow")
    aggregation_type: None

    @field_validator("aggregation_type", mode="before")
    @classmethod
    def ensure_filter_type_unknown(cls, value: Any) -> None:
        if value in set(KnownAggregationType):
            raise ValueError(
                f"{value} is an known aggregation type, but validation failed, "
                "so the response was not as expected. "
                "Try upgrading the weaviate-agents package to a new version."
            )
        return None

    def model_post_init(self, context: Any) -> None:
        warnings.warn(
            f"The aggregation_type {self.aggregation_type} wasn't recognised. "
            "Try upgrading the weaviate-agents package to a new version."
        )


PropertyAggregation = Union[
    IntegerPropertyAggregation,
    TextPropertyAggregation,
    BooleanPropertyAggregation,
    DatePropertyAggregation,
    UnknownPropertyAggregation,
]


class AggregationResult(BaseModel):
    """The aggregations to be performed on a collection in a vector database.

    They should be based on the original user query and can include multiple
    aggregations across different properties and metrics.
    """

    search_query: Optional[str] = None
    groupby_property: Optional[str] = None
    aggregations: list[PropertyAggregation]
    filters: list[PropertyFilter] = []


class AggregationResultWithCollection(AggregationResult):
    collection: str


class QueryResultWithCollection(QueryResult):
    collection: str


class Source(BaseModel):
    object_id: str
    collection: str


class QueryAgentResponse(BaseModel):
    original_query: str
    collection_names: list[str]
    searches: list[list[QueryResultWithCollection]]
    aggregations: list[list[AggregationResultWithCollection]]
    usage: Usage
    total_time: float
    aggregation_answer: Optional[str] = None
    has_aggregation_answer: bool
    has_search_answer: bool
    is_partial_answer: bool
    missing_information: list[str]
    final_answer: str
    sources: list[Source]

    def display(self) -> None:
        """Display a pretty-printed summary of the QueryAgentResponse object."""
        print_query_agent_response(self)
        return None



from typing import Any, List, Optional, Union
from uuid import UUID

import httpx
from weaviate.classes.query import Move, Rerank
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import (
    METADATA,
    HybridFusion,
    HybridVectorType,
    TargetVectorJoinType,
)
from weaviate.collections.classes.internal import ReturnProperties, ReturnReferences

from weaviate_agents.personalization.classes import (
    BM25QueryParameters,
    HybridQueryParameters,
    NearTextQueryParameters,
    PersonalizationRequest,
    PersonalizedQueryResponse,
    QueryParameters,
    QueryRequest,
)


class PersonalizedQuery:
    def __init__(
        self,
        agents_host: str,
        headers: dict,
        persona_id: UUID,
        personalization_request: PersonalizationRequest,
        timeout: Optional[int] = None,
        strength: float = 1.1,
        overfetch_factor: float = 1.5,
        recent_interactions_count: int = 100,
        decay_rate: float = 0.1,
    ):
        self._route = f"{agents_host}/personalization/query"
        self._headers = headers
        self.persona_id = persona_id
        self.timeout = timeout

        self.personalization_request = personalization_request
        self.strength = strength
        self.overfetch_factor = overfetch_factor
        self.recent_interactions_count = recent_interactions_count
        self.decay_rate = decay_rate

    def near_text(
        self,
        query: Union[List[str], str],
        *,
        certainty: Union[int, float, None] = None,
        distance: Union[int, float, None] = None,
        move_to: Optional[Move] = None,
        move_away: Optional[Move] = None,
        limit: Union[int, None] = None,
        offset: Union[int, None] = None,
        auto_limit: Union[int, None] = None,
        filters: Optional[_Filters] = None,
        # group_by,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: Union[bool, str, List[str]] = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[dict]] = None,
        return_references: Optional[ReturnReferences[dict]] = None,
    ) -> PersonalizedQueryResponse:
        query_parameters = NearTextQueryParameters(
            query=query,
            certainty=certainty,
            distance=distance,
            move_to=move_to,
            move_away=move_away,
            limit=limit,
            offset=offset,
            auto_limit=auto_limit,
            filters=filters,
            rerank=rerank,
            target_vector=target_vector,
            include_vector=include_vector,
            return_metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
        )
        return self._get_query_response(query_parameters)

    def bm25(
        self,
        query: Optional[str],
        *,
        query_properties: Optional[List[str]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        # group_by,
        rerank: Optional[Rerank] = None,
        include_vector: Union[bool, str, List[str]] = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[dict]] = None,
        return_references: Optional[ReturnReferences[dict]] = None,
    ) -> PersonalizedQueryResponse:
        query_parameters = BM25QueryParameters(
            query=query,
            query_properties=query_properties,
            limit=limit,
            offset=offset,
            auto_limit=auto_limit,
            filters=filters,
            rerank=rerank,
            include_vector=include_vector,
            return_metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
        )
        return self._get_query_response(query_parameters)

    def hybrid(
        self,
        query: Optional[str],
        *,
        alpha: Union[int, float] = 0.7,
        vector: Optional[HybridVectorType] = None,
        query_properties: Optional[List[str]] = None,
        fusion_type: Optional[HybridFusion] = None,
        max_vector_distance: Optional[Union[int, float]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        auto_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        # group_by,
        rerank: Optional[Rerank] = None,
        target_vector: Optional[TargetVectorJoinType] = None,
        include_vector: Union[bool, str, List[str]] = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[dict]] = None,
        return_references: Optional[ReturnReferences[dict]] = None,
    ) -> PersonalizedQueryResponse:
        query_parameters = HybridQueryParameters(
            query=query,
            alpha=alpha,
            vector=vector,
            query_properties=query_properties,
            fusion_type=fusion_type,
            max_vector_distance=max_vector_distance,
            limit=limit,
            offset=offset,
            auto_limit=auto_limit,
            filters=filters,
            rerank=rerank,
            target_vector=target_vector,
            include_vector=include_vector,
            return_metadata=return_metadata,
            return_properties=return_properties,
            return_references=return_references,
        )
        return self._get_query_response(query_parameters)

    def _get_request_data(self, query_parameters: QueryParameters) -> dict[str, Any]:
        query_request = QueryRequest.model_validate(
            {
                "persona_id": self.persona_id,
                "strength": self.strength,
                "recent_interactions_count": self.recent_interactions_count,
                "decay_rate": self.decay_rate,
                "overfetch_factor": self.overfetch_factor,
                "query_parameters": query_parameters,
            }
        )
        return {
            "query_request": query_request.model_dump(mode="json"),
            "personalization_request": self.personalization_request.model_dump(
                mode="json"
            ),
        }

    def _get_query_response(
        self, query_parameters: QueryParameters
    ) -> PersonalizedQueryResponse:
        response = httpx.post(
            self._route,
            headers=self._headers,
            json=self._get_request_data(query_parameters),
            timeout=self.timeout,
        )
        if response.is_error:
            raise Exception(response.text)

        return PersonalizedQueryResponse.model_validate(response.json())



from .operation import (
    AppendPropertyOperation,
    DependentOperationStep,
    Operations,
    OperationStep,
    OperationType,
    TransformationResponse,
    UpdatePropertyOperation,
)

__all__ = [
    "OperationType",
    "OperationStep",
    "AppendPropertyOperation",
    "UpdatePropertyOperation",
    "DependentOperationStep",
    "TransformationResponse",
    "Operations",
]



from weaviate_agents.personalization.classes import (
    Persona,
    PersonaInteraction,
    PersonaInteractionResponse,
    PersonalizationAgentGetObjectsResponse,
    PersonalizedObject,
    PersonalizedQueryResponse,
    Usage,
)

__all__ = [
    "Persona",
    "PersonaInteraction",
    "PersonaInteractionResponse",
    "PersonalizationAgentGetObjectsResponse",
    "PersonalizedObject",
    "PersonalizedQueryResponse",
    "Usage",
]



from ...classes.core import Usage
from .persona import Persona, PersonaInteraction, PersonaInteractionResponse
from .query import (
    BM25QueryParameters,
    HybridQueryParameters,
    NearTextQueryParameters,
    QueryParameters,
    QueryRequest,
)
from .request import GetObjectsRequest, PersonalizationRequest
from .response import (
    PersonalizationAgentGetObjectsResponse,
    PersonalizedObject,
    PersonalizedQueryResponse,
)

__all__ = [
    "Persona",
    "PersonaInteraction",
    "PersonaInteractionResponse",
    "PersonalizationAgentGetObjectsResponse",
    "PersonalizedObject",
    "Usage",
    "PersonalizedQueryResponse",
    "GetObjectsRequest",
    "PersonalizationRequest",
    "BM25QueryParameters",
    "HybridQueryParameters",
    "NearTextQueryParameters",
    "QueryParameters",
    "QueryRequest",
]



from unittest.mock import Mock
from uuid import uuid4

import httpx
import pytest
import weaviate.classes as wvc

from weaviate_agents.personalization.classes import (
    PersonalizationRequest,
    PersonalizedQueryResponse,
)
from weaviate_agents.personalization.query import PersonalizedQuery


def test_near_text(mock_httpx: Mock, personalization_request: PersonalizationRequest):
    query = PersonalizedQuery(
        agents_host="http://dummy-agent",
        headers={},
        persona_id=uuid4(),
        personalization_request=personalization_request,
    )
    output = query.near_text(
        query="Test query!",
        move_to=wvc.query.Move(force=1.5, concepts=["Towards me..."]),
        filters=wvc.query.Filter.all_of(
            [
                wvc.query.Filter.by_property("genre").equal("horror"),
                wvc.query.Filter.by_property("year").greater_or_equal("2000"),
            ]
        ),
        return_metadata=wvc.query.MetadataQuery.full(),
    )
    assert isinstance(output, PersonalizedQueryResponse)
    assert output.model_dump(mode="json") == MOCK_RESPONSE

    mock_httpx.post.assert_called_once()
    request_json = mock_httpx.post.call_args.kwargs["json"]
    request_params = request_json["query_request"]["query_parameters"]
    assert request_params["query_method"] == "near_text"


def test_bm25(mock_httpx: Mock, personalization_request: PersonalizationRequest):
    query = PersonalizedQuery(
        agents_host="http://dummy-agent",
        headers={},
        persona_id=uuid4(),
        personalization_request=personalization_request,
    )
    output = query.bm25(
        query="Test query!",
        filters=wvc.query.Filter.all_of(
            [
                wvc.query.Filter.by_property("genre").equal("horror"),
                wvc.query.Filter.by_property("year").greater_or_equal("2000"),
            ]
        ),
        return_metadata=wvc.query.MetadataQuery.full(),
    )
    assert isinstance(output, PersonalizedQueryResponse)
    assert output.model_dump(mode="json") == MOCK_RESPONSE

    mock_httpx.post.assert_called_once()
    request_json = mock_httpx.post.call_args.kwargs["json"]
    request_params = request_json["query_request"]["query_parameters"]
    assert request_params["query_method"] == "bm25"


def test_hybrid(mock_httpx: Mock, personalization_request: PersonalizationRequest):
    query = PersonalizedQuery(
        agents_host="http://dummy-agent",
        headers={},
        persona_id=uuid4(),
        personalization_request=personalization_request,
    )
    output = query.hybrid(
        query="Test query!",
        filters=wvc.query.Filter.all_of(
            [
                wvc.query.Filter.by_property("genre").equal("horror"),
                wvc.query.Filter.by_property("year").greater_or_equal("2000"),
            ]
        ),
        return_metadata=wvc.query.MetadataQuery.full(),
        vector=wvc.query.HybridVector.near_text(
            "Vector query?!",
            move_away=wvc.query.Move(force=1.5, concepts=["Not this way!"]),
        ),
    )
    assert isinstance(output, PersonalizedQueryResponse)
    assert output.model_dump(mode="json") == MOCK_RESPONSE

    mock_httpx.post.assert_called_once()
    request_json = mock_httpx.post.call_args.kwargs["json"]
    request_params = request_json["query_request"]["query_parameters"]
    assert request_params["query_method"] == "hybrid"


@pytest.fixture()
def mock_httpx(monkeypatch):
    mock = Mock()
    mock.post.return_value = httpx.Response(
        status_code=200,
        request=httpx.Request("post", "http://dummy-agent"),
        json=MOCK_RESPONSE,
    )
    monkeypatch.setattr("weaviate_agents.personalization.query.httpx", mock)
    return mock


@pytest.fixture()
def personalization_request():
    return PersonalizationRequest(
        collection_name="collection",
        create=False,
    )


MOCK_RESPONSE = {
    "objects": [
        {
            "uuid": "eba77486-3376-4d38-8fde-bfe978eaf4d4",
            "metadata": {
                "creation_time": None,
                "last_update_time": None,
                "distance": None,
                "certainty": None,
                "score": None,
                "explain_score": None,
                "is_consistent": None,
                "rerank_score": None,
            },
            "properties": {
                "title": "Avatar",
            },
            "references": None,
            "vector": {},
            "collection": "Movies",
        },
        {
            "uuid": "418691ee-241e-4671-9960-001bada94263",
            "metadata": {
                "creation_time": None,
                "last_update_time": None,
                "distance": None,
                "certainty": None,
                "score": None,
                "explain_score": None,
                "is_consistent": None,
                "rerank_score": None,
            },
            "properties": {
                "title": "Pandora",
                "vote_count": 0,
            },
            "references": None,
            "vector": {},
            "collection": "Movies",
        },
    ],
    "usage": {
        "requests": 0,
        "request_tokens": 0,
        "response_tokens": 0,
        "total_tokens": 0,
        "details": None,
    },
}



"""Tests for verifying correct package imports and public API exposure."""


def test_package_imports():
    """Test that all public modules and classes can be imported correctly.

    This test verifies that the essential modules, classes and functions from
    weaviate_agents are properly exposed and can be imported through the public API.
    """
    # Test direct imports from the package
    import weaviate_agents
    from weaviate_agents.classes.query import (
        QueryAgentCollectionConfig,
        QueryAgentResponse,
    )
    from weaviate_agents.query import QueryAgent

    # Verify the imported items are the correct types
    assert isinstance(
        QueryAgentCollectionConfig, type
    ), "QueryAgentCollection should be a class"
    assert isinstance(QueryAgentResponse, type), "QueryAgentResponse should be a class"
    assert isinstance(QueryAgent, type), "QueryAgent should be a class"

    # Optional: Test that __all__ is defined and contains expected modules
    if hasattr(weaviate_agents, "__all__"):
        expected_modules = {
            "base",
            "classes",
            "personalization",
            "query",
            "transformation",
            "utils",
        }
        assert (
            set(weaviate_agents.__all__) == expected_modules
        ), "Package __all__ should contain all public modules"


def test_class_exports():
    """Test that all data model classes are properly exported through the classes module."""
    from weaviate_agents.classes import (
        AggregationResult,
        AggregationResultWithCollection,
        AppendPropertyOperation,
        BooleanArrayPropertyFilter,
        BooleanMetrics,
        BooleanPropertyAggregation,
        BooleanPropertyFilter,
        ComparisonOperator,
        DateArrayPropertyFilter,
        DateMetrics,
        DatePropertyAggregation,
        DatePropertyFilter,
        DependentOperationStep,
        GeoPropertyFilter,
        IntegerArrayPropertyFilter,
        IntegerPropertyAggregation,
        IntegerPropertyFilter,
        NumericMetrics,
        Operations,
        OperationStep,
        OperationType,
        Persona,
        PersonaInteraction,
        PersonaInteractionResponse,
        PersonalizationAgentGetObjectsResponse,
        PersonalizedObject,
        PersonalizedQueryResponse,
        QueryAgentCollectionConfig,
        QueryAgentResponse,
        QueryResult,
        QueryResultWithCollection,
        Source,
        TextArrayPropertyFilter,
        TextMetrics,
        TextPropertyAggregation,
        TextPropertyFilter,
        UnknownPropertyAggregation,
        UnknownPropertyFilter,
        UpdatePropertyOperation,
        Usage,
    )

    # Verify all exports are classes
    classes = [
        QueryAgentCollectionConfig,
        QueryAgentResponse,
        Source,
        ComparisonOperator,
        IntegerPropertyFilter,
        TextPropertyFilter,
        BooleanPropertyFilter,
        QueryResult,
        NumericMetrics,
        TextMetrics,
        BooleanMetrics,
        IntegerPropertyAggregation,
        TextPropertyAggregation,
        BooleanPropertyAggregation,
        AggregationResult,
        AggregationResultWithCollection,
        AppendPropertyOperation,
        BooleanArrayPropertyFilter,
        BooleanMetrics,
        BooleanPropertyAggregation,
        BooleanPropertyFilter,
        ComparisonOperator,
        DateArrayPropertyFilter,
        DateMetrics,
        DatePropertyAggregation,
        DatePropertyFilter,
        DependentOperationStep,
        IntegerArrayPropertyFilter,
        IntegerPropertyAggregation,
        IntegerPropertyFilter,
        GeoPropertyFilter,
        NumericMetrics,
        Operations,
        OperationStep,
        OperationType,
        Persona,
        PersonaInteraction,
        PersonaInteractionResponse,
        PersonalizationAgentGetObjectsResponse,
        PersonalizedObject,
        PersonalizedQueryResponse,
        QueryAgentResponse,
        QueryResult,
        QueryResultWithCollection,
        Source,
        TextArrayPropertyFilter,
        TextMetrics,
        TextPropertyAggregation,
        TextPropertyFilter,
        UnknownPropertyAggregation,
        UnknownPropertyFilter,
        UpdatePropertyOperation,
        Usage,
    ]

    for cls in classes:
        assert isinstance(cls, type), f"{cls.__name__} should be a class"

    # Verify __all__ contains all exports
    import weaviate_agents.classes

    expected_exports = [
        "QueryAgentCollectionConfig",
        "QueryAgentResponse",
        "Source",
        "ComparisonOperator",
        "IntegerPropertyFilter",
        "TextPropertyFilter",
        "BooleanPropertyFilter",
        "QueryResult",
        "NumericMetrics",
        "TextMetrics",
        "BooleanMetrics",
        "IntegerPropertyAggregation",
        "TextPropertyAggregation",
        "BooleanPropertyAggregation",
        "AggregationResult",
        "Usage",
        "AggregationResultWithCollection",
        "QueryResultWithCollection",
        "OperationType",
        "OperationStep",
        "AppendPropertyOperation",
        "UpdatePropertyOperation",
        "DependentOperationStep",
        "Operations",
        "Persona",
        "PersonaInteraction",
        "PersonaInteractionResponse",
        "PersonalizationAgentGetObjectsResponse",
        "PersonalizedObject",
        "PersonalizedQueryResponse",
        "BooleanArrayPropertyFilter",
        "DateArrayPropertyFilter",
        "DateMetrics",
        "DatePropertyAggregation",
        "DatePropertyFilter",
        "IntegerArrayPropertyFilter",
        "GeoPropertyFilter",
        "TextArrayPropertyFilter",
        "UnknownPropertyAggregation",
        "UnknownPropertyFilter",
    ]

    assert hasattr(
        weaviate_agents.classes, "__all__"
    ), "Module should have __all__ defined"
    assert set(weaviate_agents.classes.__all__) == set(
        expected_exports
    ), "__all__ should contain all data model classes"



from abc import ABC, abstractmethod
from typing import Any, Coroutine, Generic, Optional, Union

import httpx
from weaviate.client import WeaviateAsyncClient, WeaviateClient

from weaviate_agents.base import ClientType, _BaseAgent
from weaviate_agents.query.classes import QueryAgentCollectionConfig, QueryAgentResponse


class _BaseQueryAgent(Generic[ClientType], _BaseAgent[ClientType], ABC):
    """An agent for executing agentic queries against Weaviate.

    Warning:
        Weaviate Agents - Query Agent is an early stage alpha product. The API is subject to
        breaking changes. Please ensure you are using the latest version of the client.

        For more information, see the [Weaviate Agents - Query Agent Docs](https://weaviate.io/developers/agents/query)
    """

    def __init__(
        self,
        client: ClientType,
        collections: Union[list[Union[str, QueryAgentCollectionConfig]], None] = None,
        agents_host: Union[str, None] = None,
        system_prompt: Union[str, None] = None,
        timeout: Union[int, None] = None,
    ):
        """Initialize the Query Agent.

        Args:
            client: The Weaviate client connected to a Weaviate Cloud cluster.
            collections: The collections to query. Will be overriden if passed in the `run` method.
            agents_host: Optional host of the agents service.
            system_prompt: Optional system prompt for the agent.
            timeout: The timeout for the request. Defaults to 60 seconds.
        """
        super().__init__(client, agents_host)

        self._collections = collections
        self._system_prompt = system_prompt

        self._timeout = 60 if timeout is None else timeout

        self.q_host = f"{self._agents_host}/agent/query"

    def _prepare_request_body(
        self,
        query: str,
        collections: Union[list[Union[str, QueryAgentCollectionConfig]], None] = None,
        context: Optional[QueryAgentResponse] = None,
    ) -> dict:
        """Prepare the request body for the query.

        Args:
            query: The natural language query string for the agent.
            collections: The collections to query. Will override any collections if passed in the constructor.
            context: Optional previous response from the agent.
        """
        collections = collections or self._collections
        if not collections:
            raise ValueError("No collections provided to the query agent.")

        return {
            "query": query,
            "collections": [
                collection
                if isinstance(collection, str)
                else collection.model_dump(mode="json")
                for collection in collections
            ],
            "headers": self._connection.additional_headers,
            "limit": 20,
            "previous_response": context.model_dump(mode="json") if context else None,
            "system_prompt": self._system_prompt,
        }

    @abstractmethod
    def run(
        self,
        query: str,
        collections: Union[list[Union[str, QueryAgentCollectionConfig]], None] = None,
        context: Optional[QueryAgentResponse] = None,
    ) -> Union[QueryAgentResponse, Coroutine[Any, Any, QueryAgentResponse]]:
        """Run the query agent. Must be implemented by subclasses."""
        pass


class QueryAgent(_BaseQueryAgent[WeaviateClient]):
    """An agent for executing agentic queries against Weaviate.

    Warning:
        Weaviate Agents - Query Agent is an early stage alpha product. The API is subject to
        breaking changes. Please ensure you are using the latest version of the client.

        For more information, see the [Weaviate Agents - Query Agent Docs](https://weaviate.io/developers/agents/query)
    """

    def run(
        self,
        query: str,
        collections: Union[list[Union[str, QueryAgentCollectionConfig]], None] = None,
        context: Optional[QueryAgentResponse] = None,
    ) -> QueryAgentResponse:
        """Run the query agent.

        Args:
            query: The natural language query string for the agent.
            collections: The collections to query. Will override any collections if passed in the constructor.
            context: Optional previous response from the agent.
        """
        request_body = self._prepare_request_body(query, collections, context)

        response = httpx.post(
            self.q_host,
            headers=self._headers,
            json=request_body,
            timeout=self._timeout,
        )

        if response.is_error:
            raise Exception(response.text)

        return QueryAgentResponse(**response.json())


class AsyncQueryAgent(_BaseQueryAgent[WeaviateAsyncClient]):
    """An agent for executing agentic queries against Weaviate.

    Warning:
        Weaviate Agents - Query Agent is an early stage alpha product. The API is subject to
        breaking changes. Please ensure you are using the latest version of the client.

        For more information, see the [Weaviate Agents - Query Agent Docs](https://weaviate.io/developers/agents/query)
    """

    async def run(
        self,
        query: str,
        collections: Union[list[Union[str, QueryAgentCollectionConfig]], None] = None,
        context: Optional[QueryAgentResponse] = None,
    ) -> QueryAgentResponse:
        """Run the query agent.

        Args:
            query: The natural language query string for the agent.
            collections: The collections to query. Will override any collections if passed in the constructor.
            context: Optional previous response from the agent.
        """
        request_body = self._prepare_request_body(query, collections, context)

        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.q_host,
                headers=self._headers,
                json=request_body,
                timeout=self._timeout,
            )

            if response.is_error:
                raise Exception(response.text)

            return QueryAgentResponse(**response.json())



from typing import Annotated

import pytest
from pydantic import BaseModel, ConfigDict
from weaviate.collections.classes.filters import Filter, _Filters
from weaviate.collections.classes.grpc import (
    HybridVector,
    HybridVectorType,
    Move,
)

from weaviate_agents.personalization.classes.query import (
    serialise_filter,
    serialise_hybrid_vector_type,
    serialise_move,
)


def test_serialise_move():
    class TestModel(BaseModel):
        model_config = ConfigDict(arbitrary_types_allowed=True)
        field: Annotated[Move, serialise_move]

    model = TestModel(field=Move(force=1, concepts=["ooh go this way"]))
    serialised = model.model_dump(mode="json")

    expect = {"field": {"force": 1, "concepts": ["ooh go this way"], "objects": None}}
    assert serialised == expect


@pytest.mark.parametrize(
    "field, expect",
    [
        (
            Filter.by_property("Property").equal("This Value"),
            {
                "field": {
                    "operator": "Equal",
                    "target": "Property",
                    "value": "This Value",
                }
            },
        ),
        (
            (
                Filter.by_property("Property").equal("This Value...")
                & Filter.by_property("Property").equal("... and this one!")
            ),
            {
                "field": {
                    "combine": "and",
                    "filters": [
                        {
                            "operator": "Equal",
                            "target": "Property",
                            "value": "This Value...",
                        },
                        {
                            "operator": "Equal",
                            "target": "Property",
                            "value": "... and this one!",
                        },
                    ],
                },
            },
        ),
        (
            (
                Filter.by_property("Property").equal("This Value...")
                | Filter.by_property("Property").equal("... or this one?")
            ),
            {
                "field": {
                    "combine": "or",
                    "filters": [
                        {
                            "operator": "Equal",
                            "target": "Property",
                            "value": "This Value...",
                        },
                        {
                            "operator": "Equal",
                            "target": "Property",
                            "value": "... or this one?",
                        },
                    ],
                },
            },
        ),
    ],
)
def test_serialise_filter(field, expect):
    class TestModel(BaseModel):
        model_config = ConfigDict(arbitrary_types_allowed=True)
        field: Annotated[_Filters, serialise_filter]

    model = TestModel(field=field)
    serialised = model.model_dump(mode="json")
    assert serialised == expect


@pytest.mark.parametrize(
    "field, expect",
    [
        (
            HybridVector.near_text(
                query="aaaa", move_to=Move(force=1, concepts=["aa"])
            ),
            {
                "field": {
                    "distance": None,
                    "certainty": None,
                    "text": "aaaa",
                    "move_to": {"force": 1.0, "objects": None, "concepts": ["aa"]},
                    "move_away": None,
                    "serialised_class": "_HybridNearText",
                }
            },
        ),
        (
            HybridVector.near_vector(vector=[1, 2, 3], distance=0.5),
            {
                "field": {
                    "serialised_class": "_HybridNearVector",
                    "vector": [1, 2, 3],
                    "distance": 0.5,
                    "certainty": None,
                }
            },
        ),
        ([1, 2, 3], {"field": [1, 2, 3]}),
        ([[1, 2, 3]], {"field": [[1, 2, 3]]}),
    ],
)
def test_serialise_hybrid_vector_type(field: HybridVectorType, expect: dict):
    class TestModel(BaseModel):
        model_config = ConfigDict(arbitrary_types_allowed=True)
        field: Annotated[HybridVectorType, serialise_hybrid_vector_type]

    model = TestModel(field=field)
    serialised = model.model_dump(mode="json")
    assert serialised == expect
